from mypy import api


imports = ""
imports += "import ophyd;"
imports += "from ophyd import flyers, sim;"
imports += "hw = sim.hw();"


def run(cmd):
    normal_report, error_report, exit_status = api.run(
        ["--command", cmd, "--follow-imports=skip"]
    )
    print(cmd.split(";")[-1])
    print("  ", normal_report)
    print("  ", error_report)
    assert exit_status == 0


def test_checkable():
    cmd = imports + "from bluesky.protocols import Checkable;"
    run(cmd + "foo: Checkable = hw.motor1")
    run(cmd + "foo: Checkable = ophyd.Device(name='test')")


def test_flyable():
    cmd = imports + "from bluesky.protocols import Flyable;"
    run(cmd + "foo: Flyable = hw.flyer1")
    run(cmd + "foo: Flyable = sim.TrivialFlyer()")


def test_hinted():
    cmd = imports + "from bluesky.protocols import Hinted;"
    run(cmd + "foo: Hinted = ophyd.Signal(name='test')")


def test_movable():
    cmd = imports + "from bluesky.protocols import Movable;"
    run(cmd + "foo: Movable = hw.motor1")
    run(cmd + "foo: Movable = ophyd.Device(name='test')")
    run(cmd + "foo: Movable = hw.flyer1")


def test_pausable():
    cmd = imports + "from bluesky.protocols import Pausable;"
    run(cmd + "foo: Pausable = hw.motor1")
    run(cmd + "foo: Pausable = ophyd.Device(name='test')")


def test_readable():
    cmd = imports + "from bluesky.protocols import Readable;"
    run(cmd + "foo: Readable = hw.motor1")
    run(cmd + "foo: Readable = ophyd.Device(name='test')")
    run(cmd + "foo: Readable = hw.signal1")


def test_stageable():
    cmd = imports + "from bluesky.protocols import Stageable;"
    run(cmd + "foo: Stageable = hw.motor1")
    run(cmd + "foo: Stageable = ophyd.Device(name='test')")


def test_status():
    cmd = imports + "from bluesky.protocols import Status;"
    run(cmd + "foo: Status = ophyd.status.Status()")
    run(cmd + "foo: Status = ophyd.status.StatusBase()")


def test_stoppable():
    cmd = imports + "from bluesky.protocols import Stoppable;"
    run(cmd + "foo: Stoppable = hw.motor1")
    run(cmd + "foo: Stoppable = hw.flyer1")


def test_subscribable():
    cmd = imports + "from bluesky.protocols import Subscribable;"
    run(cmd + "foo: Subscribable = hw.signal1")
    run(cmd + "foo: Subscribable = ophyd.Signal(name='test')")
    run(cmd + "foo: Subscribable = ophyd.Device(name='test')")
    run(cmd + "foo: Subscribable = hw.motor1")
    run(cmd + "foo: Subscribable = hw.flyer1")


if __name__ == "__main__":
    test_checkable()
    test_flyable()
    test_hinted()
    test_movable()
    test_pausable()
    test_readable()
    test_stageable()
    test_status()
    test_stoppable()
    test_subscribable()



import pytest

from ophyd.docs import get_device_info


def test_get_device_info_scaler():
    # `get_device_info` done by name, due to how sphinx directives work:
    info = get_device_info("ophyd.scaler", "EpicsScaler")
    assert len(info)

    # A list of component information dictionaries:
    components_by_attr = {item["attr"]: item for item in info}
    assert "count_mode" in components_by_attr
    assert "channels (DDC)" in components_by_attr

    # And detailed nested component information for a DDC:
    nested = components_by_attr["channels (DDC)"]["nested_components"]
    nested_components_by_attr = {item["attr"]: item for item in nested}
    assert "chan1" in nested_components_by_attr


@pytest.mark.parametrize(
    "module, class_name",
    [
        pytest.param("ophyd.areadetector.base", "ADBase"),
        pytest.param("ophyd.areadetector.cam", "PcoDetectorIO"),
        pytest.param("ophyd.areadetector.cam", "PcoDetectorSimIO"),
        pytest.param("ophyd.areadetector.common_plugins", "CommonAttributePlugin"),
        pytest.param("ophyd.areadetector.common_plugins", "CommonGatherPlugin"),
        pytest.param("ophyd.areadetector.common_plugins", "CommonOverlayPlugin"),
        pytest.param("ophyd.areadetector.common_plugins", "CommonPlugins"),
        pytest.param("ophyd.areadetector.common_plugins", "CommonROIStatPlugin"),
        pytest.param("ophyd.areadetector.detectors", "DetectorBase"),
        pytest.param("ophyd.areadetector.plugins", "AttrPlotPlugin"),
        pytest.param("ophyd.areadetector.plugins", "AttributeNPlugin"),
        pytest.param("ophyd.areadetector.plugins", "AttributePlugin"),
        pytest.param("ophyd.areadetector.plugins", "CircularBuffPlugin"),
        pytest.param("ophyd.areadetector.plugins", "CodecPlugin"),
        pytest.param("ophyd.areadetector.plugins", "FFTPlugin"),
        pytest.param("ophyd.areadetector.plugins", "GatherNPlugin"),
        pytest.param("ophyd.areadetector.plugins", "Overlay"),
        pytest.param("ophyd.areadetector.plugins", "PluginBase"),
        pytest.param("ophyd.areadetector.plugins", "PosPlugin"),
        pytest.param("ophyd.areadetector.plugins", "PvaPlugin"),
        pytest.param("ophyd.areadetector.plugins", "ROIStatNPlugin"),
        pytest.param("ophyd.areadetector.plugins", "ROIStatPlugin"),
        pytest.param("ophyd.areadetector.plugins", "ScatterPlugin"),
        pytest.param("ophyd.areadetector.plugins", "TimeSeriesNPlugin"),
        pytest.param("ophyd.areadetector.plugins", "TimeSeriesPlugin"),
        pytest.param("ophyd.epics_motor", "EpicsMotor"),
        pytest.param("ophyd.epics_motor", "MotorBundle"),
        pytest.param("ophyd.flyers", "AreaDetectorTimeseriesCollector"),
        pytest.param("ophyd.flyers", "WaveformCollector"),
        pytest.param("ophyd.mca", "EpicsDXP"),
        pytest.param("ophyd.mca", "EpicsDXPBaseSystem"),
        pytest.param("ophyd.mca", "EpicsDXPLowLevel"),
        pytest.param("ophyd.mca", "EpicsDXPLowLevelParameter"),
        pytest.param("ophyd.mca", "EpicsDXPMapping"),
        pytest.param("ophyd.mca", "EpicsMCACallback"),
        pytest.param("ophyd.mca", "EpicsMCARecord"),
        pytest.param("ophyd.mca", "ROI"),
        pytest.param("ophyd.mca", "SoftDXPTrigger"),
        pytest.param("ophyd.pseudopos", "PseudoPositioner"),
        pytest.param("ophyd.pseudopos", "PseudoSingle"),
        pytest.param("ophyd.pv_positioner", "PVPositioner"),
        pytest.param("ophyd.quadem", "QuadEMPort"),
        pytest.param("ophyd.scaler", "EpicsScaler"),
        pytest.param("ophyd.scaler", "ScalerCH"),
        pytest.param("ophyd.scaler", "ScalerChannel"),
        pytest.param("ophyd.sim", "ABDetector"),
        pytest.param("ophyd.sim", "DetWithConf"),
        pytest.param("ophyd.sim", "DetWithCountTime"),
        pytest.param("ophyd.sim", "DirectImage"),
        pytest.param("ophyd.sim", "Syn2DGauss"),
        pytest.param("ophyd.sim", "SynAxis"),
        pytest.param("ophyd.sim", "SynGauss"),
        pytest.param("ophyd.tests.test_sim", "Sample"),
        pytest.param("ophyd.tests.test_sim", "SampleNested"),
    ],
)
def test_get_device_info_smoke(module, class_name):
    get_device_info(module, class_name)



import time
import logging

from .signal import EpicsSignal
from .positioner import PositionerBase
from .status import MoveStatus, wait as status_wait


logger = logging.getLogger(__name__)


class SignalPositionerMixin(PositionerBase):
    """Mixin to make a Signal a Positioner

    Should be mixed in first, with the Signal second, such that:
        set, move will be from PositionerBase

    Parameters
    ----------
    set_func : callable
        The set() functionality for the class. Must be specified as the mixin
        takes over set() functionality.
    readback_event : str, optional
        Readback value subscription event_type
    egu : str, optional
        Engineering units of positioner
    hold_on_stop : bool, optional
        When stop is called on the positioner
    """

    def __init__(
        self,
        *args,
        set_func,
        readback_event="value",
        egu="",
        hold_on_stop=False,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._egu = egu
        self._hold_on_stop = hold_on_stop
        self._internal_status = None
        self._external_status = None

        # bind method to this class instance:
        self._mixed_set = set_func.__get__(self, self.__class__)

        self.subscribe(self._position_updated, event_type=readback_event)

    def _position_updated(self, value=None, **kwargs):
        # for readback subscriptions
        self._set_position(value)

    @property
    def position(self):
        """The current position of the motor in its engineering units

        Returns
        -------
        position : any
        """
        return self.get()

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return self._egu

    def move(self, position, wait=True, moved_cb=None, timeout=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        wait : bool, optional
            Wait until motion has completed before returning
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : Status

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        if timeout is None:
            timeout = self._timeout

        self.check_value(position)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=False)
        self._reset_sub(self._SUB_REQ_DONE)

        self._moving = True
        self._run_subs(sub_type=self.SUB_START, timestamp=time.time())

        def finished(status):
            success = self._internal_status.success
            self._moving = False
            self._done_moving(success=success)

            if moved_cb is not None:
                try:
                    moved_cb(obj=self)
                except Exception as ex:
                    logger.error("Move callback failed", exc_info=ex)

            try:
                self._external_status._finished(success=success)
            except Exception as ex:
                logger.error("Status completion failed", exc_info=ex)

            self._internal_status = None
            self._external_status = None

        # this external status object is fully dependent on the internal status
        # object and does not have its own timeout/settle_time settings:
        self._external_status = MoveStatus(
            self, target=position, settle_time=0.0, timeout=None
        )

        # set() functionality depends on the signal
        self._internal_status = self._mixed_set(
            position, timeout=timeout, settle_time=self.settle_time
        )
        self._internal_status.add_callback(finished)

        if wait:
            try:
                status_wait(self._external_status)
            except RuntimeError:
                raise RuntimeError("Motion did not complete successfully")

        return self._external_status

    def stop(self, *, success=False):
        """Stops motion"""
        if self._hold_on_stop:
            self.move(self.get(), wait=False)
        # TODO status object?
        return super().stop(success=success)

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("egu", self.egu)
        yield ("hold_on_stop", self._hold_on_stop)


class EpicsSignalPositioner(SignalPositionerMixin, EpicsSignal):
    def __init__(self, read_pv, **kwargs):
        super().__init__(read_pv=read_pv, set_func=EpicsSignal.set, **kwargs)



import logging

from .utils import DisconnectedError

logger = logging.getLogger(__name__)


class UidPublish:
    """Publishes run start UID of most recently begun run to a given signal

    Processed on every start/end document.

    Note: If used with an EpicsSignal, it's recommended to use a waveform in
    place of a stringin record on the EPICS side, as the start document UID
    will be published both on run start and run completion. A stringin record
    will only process on change and monitor events will only be received on run
    start.

    Sample EPICS record definition:

        record(waveform, "$(Sys)$(Dev)UID-I") {
            # Using waveform here as it always reprocesses, so you'll get a
            # monitor event on start/stop of the run
            field(DESC, "Last run UID")
            field(FTVL, "STRING")
            field(MPST, "Always")
            info(autosaveFields_pass0, "VAL")
        }

    Parameters
    ----------
    uid_signal : Signal
        The signal to publish to
    raise_if_disconnected : bool, optional
        Fail if the UID signal is disconnected
    put_kw : kwargs, optional
        Keyword arguments to send to uid_signal.put()
    """

    def __init__(self, signal, raise_if_disconnected=False, **put_kw):
        self._uid = None
        self.last_start = None
        self.uid_signal = signal
        self.put_kw = put_kw
        self.raise_if_disconnected = raise_if_disconnected

    @property
    def uid(self):
        """The uid of the last run"""
        return self._uid

    @uid.setter
    def uid(self, uid):
        self._uid = uid

        if uid is None:
            uid = ""

        try:
            self.uid_signal.put(uid, **self.put_kw)
        except (DisconnectedError, TimeoutError):
            logger.error("UID signal disconnected. Is the IOC running?")
            if self.raise_if_disconnected:
                raise

    def clear(self):
        """Clear the run uid"""
        self.uid = None

    def __call__(self, name, doc):
        """Bluesky callback with document info"""
        if name == "start":
            self.last_start = doc

        if self.last_start and name in ("start", "stop"):
            self.uid = self.last_start["uid"]


class LastUidPublish(UidPublish):
    """Publishes run start UID of most recently completed run to a given signal

    Processed on every stop document.

    Note: If used with an EpicsSignal, it's recommended to use a waveform in
    place of a stringin record on the EPICS side, as the start document UID
    will be published both on run start and run completion. A stringin record
    will only process on change and monitor events will only be received on run
    start.

    Sample EPICS record definition:

        record(waveform, "$(Sys)$(Dev)UID-I") {
            # Using waveform here as it always reprocesses, so you'll get a
            # monitor event on start/stop of the run
            field(DESC, "Last run UID")
            field(FTVL, "STRING")
            field(MPST, "Always")
            info(autosaveFields_pass0, "VAL")
        }

    Parameters
    ----------
    uid_signal : Signal
        The signal to publish to
    raise_if_disconnected : bool, optional
        Fail if the UID signal is disconnected
    put_kw : kwargs, optional
        Keyword arguments to send to uid_signal.put()
    """

    def __call__(self, name, doc):
        """Bluesky callback with document info"""
        if name == "start":
            self.last_start = doc

        if self.last_start is not None and name == "stop":
            self.uid = self.last_start["uid"]



#!/usr/bin/env python
import os
import sys

from setuptools import find_packages, setup

import versioneer

# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 8)
if sys.version_info < min_version:
    error = """
ophyd does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(
        *(sys.version_info[:2] + min_version)
    )
    sys.exit(error)


here = os.path.abspath(os.path.dirname(__file__))

# Get the requirements from requirements.txt
with open(os.path.join(here, "requirements.txt"), "rt") as f:
    requirements = f.read().split()

# Get the long description from the README file
with open(os.path.join(here, "README.rst"), "rt", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="ophyd",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    long_description=long_description,
    license="BSD",
    python_requires=">={}".format(".".join(str(n) for n in min_version)),
    install_requires=requirements,
    packages=find_packages(),
    entry_points={
        "databroker.handlers": [
            "NPY_SEQ = ophyd.sim:NumpySeqHandler",
        ]
    },
    include_package_data=True,
    package_data={
        # NOTE: this is required in addition to MANIFEST.in, as that only
        # applies to source distributions
        # Include our documentation helpers:
        "": ["*.rst"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



"""Mixin classes that customize the triggering behavior of AreaDetector

To be used like so ::

    from ophyd.areadetector.detectors import PerkinElmerDetector
    from ophyd.areadetector.trigger_mixins import SingleTrigger

    class MyDetector(PerkinElmerDetector, SingleTrigger):
        pass
"""

import time as ttime
import logging
import itertools

from ..status import DeviceStatus
from ..device import BlueskyInterface, Staged

logger = logging.getLogger(__name__)


class ADTriggerStatus(DeviceStatus):
    """
    A Status for AreaDetector triggers

    A special status object that notifies watches (progress bars)
    based on comparing device.cam.array_counter to  device.cam.num_images.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_ts = ttime.time()

        # Notify watchers (things like progress bars) of new values
        # at the device's natural update rate.
        if not self.done:
            self.device.cam.array_counter.subscribe(self._notify_watchers)
            # some state needed only by self._notify_watchers
            self._name = self.device.name
            self._initial_count = self.device.cam.array_counter.get()
            self._target_count = self.device.cam.num_images.get()

    def watch(self, func):
        self._watchers.append(func)

    def _notify_watchers(self, value, *args, **kwargs):
        # *args and **kwargs catch extra inputs from pyepics, not needed here
        if self.done:
            self.device.cam.array_counter.clear_sub(self._notify_watchers)
        if not self._watchers:
            return
        # Always start progress bar at 0 regardless of starting value of
        # array_counter.
        current = value - self._initial_count
        target = self._target_count
        initial = 0
        time_elapsed = ttime.time() - self.start_ts
        try:
            fraction = (current - initial) / (target - initial)
        except ZeroDivisionError:
            fraction = 1
        except Exception:
            fraction = None
            time_remaining = None
        else:
            time_remaining = time_elapsed / fraction
        for watcher in self._watchers:
            watcher(
                name=self._name,
                current=current,
                initial=initial,
                target=target,
                unit="images",
                precision=0,
                fraction=fraction,
                time_elapsed=time_elapsed,
                time_remaining=time_remaining,
            )


class TriggerBase(BlueskyInterface):
    """Base class for trigger mixin classes

    Subclasses must define a method with this signature:

    `acquire_changed(self, value=None, old_value=None, **kwargs)`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # settings
        # careful here: quadEM devices have areadetector components but,
        # they have no 'cam' plugin. See QuadEM initializer.
        if hasattr(self, "cam"):
            self.stage_sigs.update(
                [
                    ("cam.acquire", 0),  # If acquiring, stop
                    ("cam.image_mode", 1),  # 'Multiple' mode
                ]
            )
            self._acquisition_signal = self.cam.acquire

        self._status = None


class SingleTrigger(TriggerBase):
    """
    This trigger mixin class takes one acquisition per trigger.

    Examples
    --------

    >>> class SimDetector(SingleTrigger):
    ...     pass
    >>> det = SimDetector('..pv..')
    # optionally, customize name of image
    >>> det = SimDetector('..pv..', image_name='fast_detector_image')
    """

    _status_type = ADTriggerStatus

    def __init__(self, *args, image_name=None, **kwargs):
        super().__init__(*args, **kwargs)
        if image_name is None:
            image_name = "_".join([self.name, "image"])
        self._image_name = image_name

    def stage(self):
        self._acquisition_signal.subscribe(self._acquire_changed)
        super().stage()

    def unstage(self):
        super().unstage()
        self._acquisition_signal.clear_sub(self._acquire_changed)

    def trigger(self):
        "Trigger one acquisition."
        if self._staged != Staged.yes:
            raise RuntimeError(
                "This detector is not ready to trigger."
                "Call the stage() method before triggering."
            )

        self._status = self._status_type(self)
        self._acquisition_signal.put(1, wait=False)
        self.generate_datum(self._image_name, ttime.time(), {})
        return self._status

    def _acquire_changed(self, value=None, old_value=None, **kwargs):
        "This is called when the 'acquire' signal changes."
        if self._status is None:
            return
        if (old_value == 1) and (value == 0):
            # Negative-going edge means an acquisition just finished.
            self._status.set_finished()
            self._status = None


class MultiTrigger(TriggerBase):
    """This trigger mixin class can take multiple acquisitions per trigger.

    This can be used to give more control to the detector. One call to
    'trigger' can be interpreted by the detector as a call to take several
    acquisitions with, for example, different gain settings or shutter
    positions.

    The are two levels of nesting here:

     - cycling through different actions on successive calls to `trigger`
     - within each trigger, executing a list of acquisitions with different
       settings

    See the example below, which takes and 3 and 1 acquisitions in
    alternation.

    Examples
    --------

    >>> class MyDetector(SimDetector, MultiTrigger):
    ...     pass
    # EXAMPLE:
    # 1. On the first trigger, close the shutter and acquire three images
    # with different gain settings on the detector. Then open the shutter
    # and take a light frame.
    # 2. On the next trigger, just take a light frame.
    # Repeat.
    #
    # Each element of this list specifies one acquisition. It gives a
    # a label for each kind of image that will be taken and a dictionary
    # mapping signals to values that must be set for that acquisition.
    >>> dark_and_light = [('gain1', {'shutter': 'close', 'image_gain': 1}),
    ...                   ('gain2', {'image_gain': 2}),
    ...                   ('gain8', {'image_gain': 8}),
    ...                   ('light', {'shutter': 'open'})],
    # This list only has one element; it will only take one acquisition.
    >>> light_only = [('light', {'shutter': 'open'}]]
    # Finally, put the lists together. The detector will cycle through
    # this list as it is triggered.
    >>> trigger_cycle = [dark_and_light, light_only]
    >>> det = MyDetector(trigger_cycle=trigger_cycle)
    # Note: for simplicity, the settings were specified as dictionaries. If
    # you need to control the order that they are processed, use
    # OrderedDict instead.
    """

    # OphydObj subscriptions
    _SUB_ACQ_DONE = "acq_done"

    def __init__(self, *args, trigger_cycle=None, **kwargs):
        if trigger_cycle is None:
            raise ValueError("must provide trigger_cycle -- see docstring")
        self.trigger_cycle = trigger_cycle
        super().__init__(*args, **kwargs)

    def stage(self):
        self._acquisition_signal.subscribe(self._acquire_changed)
        super().stage()

    def unstage(self):
        super().unstage()
        self._acquisition_signal.clear_sub(self._acquire_changed)

    @property
    def trigger_cycle(self):
        return self._trigger_cycle

    @trigger_cycle.setter
    def trigger_cycle(self, val):
        self._trigger_cycle = itertools.cycle(val)

    def trigger(self):
        "Trigger one or more acquisitions."
        if self._staged != Staged.yes:
            raise RuntimeError(
                "This detector is not ready to trigger."
                "Call the stage() method before triggering."
            )

        # For each trigger, we have a list of one of more acquisitions to
        # take. These are names (e.g., 'light' or 'dark') paired with
        # an ordered dict of signals and values to set.
        acq_list = next(self.trigger_cycle)
        self._acq_iter = iter(acq_list)

        # GET READY...

        # Reset subscritpions.
        self._reset_sub(self._SUB_ACQ_DONE)

        # When each acquisition finishes, it will immedately start the next
        # one until the desired number has been taken.
        self.subscribe(self._acquire, event_type=self._SUB_ACQ_DONE, run=False)

        # When *all* the acquisitions are done, increment the trigger counter
        # and kick the status object.
        self._status = DeviceStatus(self)

        # GO!
        self._acquire()
        return self._status

    def _acquire(self, **kwargs):
        "Start the next acquisition or find that all acquisitions are done."
        try:
            key, signals_settings = next(self._acq_iter)
        except StopIteration:
            logger.debug("Trigger cycle is complete.")
            self._status.set_finished()
            return
        logger.debug("Configuring signals for acquisition labeled %r", key)
        for sig, val in signals_settings.items():
            sig.set(val).wait()
        self.generate_datum(key, ttime.time(), {})
        self._acquisition_signal.put(1, wait=False)

    def _acquire_changed(self, value=None, old_value=None, **kwargs):
        "This is called when the 'acquire' signal changes."
        logger.debug(
            "_acquire_chaged has been called: old_value %r, value %r", old_value, value
        )
        if self._status is None:
            return
        if (old_value == 1) and (value == 0):
            # Negative-going edge means an acquisition just finished.
            self._run_subs(sub_type=self._SUB_ACQ_DONE)



from ophyd import Device, Component, Signal, MotorBundle, Kind
from ophyd.sim import SynAxis


def test_device_hints():
    # Default hints is 'fields' with an empty list.
    assert {"fields": []} == Device("", name="dev").hints

    # Class-provided default works.
    class Dongle(Device):
        a = Component(Signal, kind=Kind.hinted)
        b = Component(Signal)

    assert {"fields": ["dev_a"]} == Dongle(name="dev").hints
    assert {"fields": ["dev_a"]} == Dongle(name="dev").hints


def test_motor_bundle_hints():
    class Bundle(MotorBundle):
        a = Component(SynAxis, kind=Kind.hinted)
        b = Component(SynAxis, kind=Kind.hinted)

    assert {"fields": ["dev_a", "dev_b"]} == Bundle(name="dev").hints


def test_pseudopos_hints(hw):
    assert len(hw.pseudo3x3.hints["fields"]) == 3
    assert len(hw.pseudo1x3.hints["fields"]) == 1


def test_hints_are_lazy():
    class Stuff(Device):
        sig = Component(Signal, lazy=True)

    class Thing(Device):
        hinted = Component(Stuff, kind=Kind.hinted, lazy=True)
        normal = Component(Stuff, lazy=True)
        config = Component(Stuff, kind=Kind.config, lazy=True)
        omitted = Component(Stuff, kind=Kind.omitted, lazy=True)

    class ThingHaver(Device):
        t = Component(Thing, lazy=True)

    th = ThingHaver(name="t")

    # Lazy Component not instantiated.
    assert not th._signals
    # This instantiates only the things it needs to.
    th.hints
    assert "t" in th._signals
    assert "hinted" in th.t._signals
    assert "normal" in th.t._signals
    assert "config" not in th.t._signals
    assert "omitted" not in th.t._signals



import logging
import warnings

logger = logging.getLogger(__name__)

__all__ = ["setup", "logger"]


def setup():
    from .. import get_cl

    warnings.warn(
        "This function is deprecated in ophyd 1.2 "
        "and will be removed in ophyd 1.3. "
        "setup is now automatically called."
    )
    return get_cl().setup(logger)



import time
import logging
import pytest
from ophyd import EpicsSignal, EpicsSignalRO

from numpy.testing import assert_almost_equal


logger = logging.getLogger(__name__)


@pytest.mark.motorsim
def test_read_pv_timestamp_no_monitor(motor):
    motor.set(0, wait=True)
    sp = EpicsSignal(motor.user_setpoint.pvname, name="test")
    rbv = EpicsSignalRO(motor.user_readback.pvname, name="test")

    assert rbv.get() == sp.get()
    rbv_value0 = rbv.get()
    ts0 = rbv.timestamp
    sp.put(sp.get() + 0.1)
    time.sleep(0.5)
    rbv_value1 = rbv.get()
    ts1 = rbv.timestamp
    assert ts1 > ts0
    assert_almost_equal(rbv_value0 + 0.1, rbv_value1)

    sp.put(sp.get() - 0.1)


@pytest.mark.motorsim
def test_write_pv_timestamp_no_monitor(motor):
    motor.set(0, wait=True)
    sp = EpicsSignal(motor.user_setpoint.pvname, name="test")

    sp_value0 = sp.get()
    ts0 = sp.timestamp
    sp.put(sp_value0 + 0.1, wait=True)
    time.sleep(0.1)

    sp_value1 = sp.get()
    ts1 = sp.timestamp
    assert ts1 > ts0
    assert_almost_equal(sp_value0 + 0.1, sp_value1)

    sp.put(sp.get() - 0.1, wait=True)


@pytest.mark.motorsim
def test_read_pv_timestamp_monitor(motor):
    motor.set(0, wait=True)

    sp = EpicsSignal(motor.user_setpoint.pvname, auto_monitor=True, name="test")
    rbv = EpicsSignalRO(motor.user_readback.pvname, auto_monitor=True, name="test")

    rbv_value0 = rbv.get()
    ts0 = rbv.timestamp
    sp.put(rbv_value0 + 0.1, wait=True)
    time.sleep(0.2)

    rbv_value1 = rbv.value
    ts1 = rbv.timestamp
    assert ts1 > ts0
    assert_almost_equal(rbv_value0 + 0.1, rbv_value1)

    sp.put(sp.value - 0.1, wait=True)


@pytest.mark.motorsim
def test_write_pv_timestamp_monitor(motor):
    motor.set(0, wait=True)
    sp = EpicsSignal(motor.user_setpoint.pvname, auto_monitor=True, name="test")

    sp_value0 = sp.get()
    ts0 = sp.timestamp
    sp.put(sp_value0 + 0.1, wait=True)
    time.sleep(0.1)

    sp_value1 = sp.value
    ts1 = sp.timestamp
    assert ts1 > ts0
    assert_almost_equal(sp_value0 + 0.1, sp_value1)

    sp.put(sp.value - 0.1, wait=True)



import datetime
import logging
import os
import shutil
import time
from unittest.mock import Mock

import numpy as np
import pytest
from io import StringIO
from pathlib import PurePath, Path

from ophyd.utils.paths import make_dir_tree
from ophyd import (
    SimDetector,
    SingleTrigger,
    Component,
    Device,
    DynamicDeviceComponent,
    Kind,
    wait,
)
from ophyd.areadetector.plugins import (
    ImagePlugin,
    StatsPlugin,
    ColorConvPlugin,
    ProcessPlugin,
    OverlayPlugin,
    ROIPlugin,
    TransformPlugin,
    NetCDFPlugin,
    TIFFPlugin,
    JPEGPlugin,
    HDF5Plugin,
    # FilePlugin
)
from ophyd.areadetector.base import NDDerivedSignal
from ophyd.areadetector.filestore_mixins import (
    FileStoreTIFF,
    FileStoreIterativeWrite,
    FileStoreHDF5,
)

# we do not have nexus installed on our test IOC
# from ophyd.areadetector.plugins import NexusPlugin
from ophyd.areadetector.plugins import PluginBase
from ophyd.areadetector.util import stub_templates
from ophyd.device import (
    Component as Cpt,
)
from ophyd.signal import Signal
import uuid

logger = logging.getLogger(__name__)
ad_path = "/epics/support/areaDetector/1-9-1/ADApp/Db/"


class DummyFS:
    def __init__(self):
        self.resource = {}
        self.datum = {}
        self.datum_by_resource = {}

    def insert_resource(
        self,
        spec,
        resource_path,
        resource_kwargs,
        root=None,
        path_semantics="posix",
        uid=None,
        run_start=None,
        ignore_duplicate_error=False,
    ):
        self.resource[uid] = {
            "spec": spec,
            "resource_path": resource_path,
            "root": root,
            "resource_kwargs": resource_kwargs,
            "uid": uid,
            "path_semantics": path_semantics,
        }
        self.datum_by_resource[uid] = []
        return uid

    def register_resource(
        self, spec, root, rpath, rkwargs, path_semantics="posix", run_start=None
    ):

        uid = str(uuid.uuid4())
        return self.insert_resource(
            spec,
            rpath,
            rkwargs,
            root=root,
            path_semantics=path_semantics,
            uid=uid,
            run_start=run_start,
        )

    def register_datum(self, resource_uid, datum_kwargs):
        datum_id = str(uuid.uuid4())
        return self.insert_datum(resource_uid, datum_id, datum_kwargs)

    def insert_datum(
        self, resource, datum_id, datum_kwargs, ignore_duplicate_error=False
    ):
        datum = {
            "resource": resource,
            "datum_id": datum_id,
            "datum_kwargs": datum_kwargs,
        }
        self.datum_by_resource[resource].append(datum)
        self.datum[datum_id] = datum
        return datum_id


# lifted from soft-matter/pims source
def _recursive_subclasses(cls):
    "Return all subclasses (and their subclasses, etc.)."
    # Source: http://stackoverflow.com/a/3862957/1221924
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in _recursive_subclasses(s)
    ]


@pytest.mark.adsim
def test_basic(cleanup, ad_prefix):
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")

    det = MyDetector(ad_prefix, name="test")
    print(det.tiff1.plugin_type)
    cleanup.add(det)

    det.wait_for_connection()

    det.cam.acquire_time.put(0.5)
    det.cam.acquire_period.put(0.5)
    det.cam.num_images.put(1)
    det.cam.image_mode.put(det.cam.ImageMode.SINGLE)
    det.stage()
    st = det.trigger()
    wait(st, timeout=5)
    det.unstage()


@pytest.mark.skipif("not os.path.exists(ad_path)")
def test_stubbing():
    try:
        for line in stub_templates(ad_path):
            logger.debug("Stub line: %s", line)
    except OSError:
        # self.fail('AreaDetector db path needed to run test')
        pass


@pytest.mark.adsim
def test_detector(ad_prefix, cleanup):
    det = SimDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.find_signal("a", f=StringIO())
    det.find_signal("a", use_re=True, f=StringIO())
    det.find_signal("a", case_sensitive=True, f=StringIO())
    det.find_signal("a", use_re=True, case_sensitive=True, f=StringIO())
    det.component_names
    det.report

    cam = det.cam

    cam.image_mode.put("Single")
    # plugins don't live on detectors now:
    # det.image1.enable.put('Enable')
    cam.array_callbacks.put("Enable")

    st = det.trigger()
    repr(st)
    det.read()

    # values = tuple(det.gain_xy.get())
    cam.gain_xy.put(cam.gain_xy.get(), wait=True)

    # fail when only specifying x
    with pytest.raises(ValueError):
        cam.gain_xy.put((0.0,), wait=True)

    det.describe()
    det.report


@pytest.mark.adsim
def test_tiff_plugin(ad_prefix, cleanup):
    # det = AreaDetector(ad_prefix)
    class TestDet(SimDetector):
        p = Cpt(TIFFPlugin, "TIFF1:")

    det = TestDet(ad_prefix, name="test")
    cleanup.add(det)

    plugin = det.p

    plugin.file_template.put("%s%s_%3.3d.tif")

    plugin.array_pixels
    plugin


@pytest.mark.adsim
def test_hdf5_plugin(ad_prefix, cleanup):
    class MyDet(SimDetector):
        p = Cpt(HDF5Plugin, suffix="HDF1:")

    d = MyDet(ad_prefix, name="d")
    cleanup.add(d)

    d.p.file_path.put("/tmp")
    d.p.file_name.put("--")
    d.p.warmup()
    d.stage()
    print(d.p.read_configuration())
    d.p.describe_configuration()
    d.unstage()


@pytest.mark.adsim
def test_subclass(ad_prefix, cleanup):
    class MyDetector(SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)
    det.wait_for_connection()

    print(det.describe())
    print(det.tiff1.capture.describe())


@pytest.mark.adsim
def test_getattr(ad_prefix, cleanup):
    class MyDetector(SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)
    assert getattr(det, "tiff1.name") == det.tiff1.name
    assert getattr(det, "tiff1") is det.tiff1
    # raise
    # TODO subclassing issue


@pytest.mark.adsim
def test_invalid_plugins(ad_prefix, cleanup):
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")
        stats1 = Cpt(StatsPlugin, "Stats1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.wait_for_connection()
    det.tiff1.nd_array_port.put(det.cam.port_name.get())
    det.stats1.nd_array_port.put("AARDVARK")

    with pytest.raises(RuntimeError):
        det.stage()

    with pytest.raises(RuntimeError):
        det.validate_asyn_ports()

    assert ["AARDVARK"] == det.missing_plugins()


@pytest.mark.adsim
def test_validate_plugins_no_portname(ad_prefix, cleanup):
    class MyDetector(SingleTrigger, SimDetector):
        roi1 = Cpt(ROIPlugin, "ROI1:")
        over1 = Cpt(OverlayPlugin, "Over1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.roi1.nd_array_port.put(det.cam.port_name.get())
    det.over1.nd_array_port.put(det.roi1.port_name.get())

    det.validate_asyn_ports()


@pytest.mark.adsim
def test_get_plugin_by_asyn_port(ad_prefix, cleanup):
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")
        stats1 = Cpt(StatsPlugin, "Stats1:")
        roi1 = Cpt(ROIPlugin, "ROI1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.tiff1.nd_array_port.put(det.cam.port_name.get())
    det.roi1.nd_array_port.put(det.cam.port_name.get())
    det.stats1.nd_array_port.put(det.roi1.port_name.get())

    det.validate_asyn_ports()

    assert det.tiff1 is det.get_plugin_by_asyn_port(det.tiff1.port_name.get())
    assert det.cam is det.get_plugin_by_asyn_port(det.cam.port_name.get())
    assert det.roi1 is det.get_plugin_by_asyn_port(det.roi1.port_name.get())


@pytest.mark.adsim
def test_get_plugin_by_asyn_port_nested(ad_prefix, cleanup):
    # Support nested plugins
    class PluginGroup(Device):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")

    class MyDetector(SingleTrigger, SimDetector):
        plugins = Cpt(PluginGroup, "")
        roi1 = Cpt(ROIPlugin, "ROI1:")
        stats1 = Cpt(StatsPlugin, "Stats1:")

    nested_det = MyDetector(ad_prefix, name="nested_test")
    cleanup.add(nested_det)

    nested_det.stats1.nd_array_port.put(nested_det.roi1.port_name.get())
    nested_det.plugins.tiff1.nd_array_port.put(nested_det.cam.port_name.get())
    nested_det.roi1.nd_array_port.put(nested_det.cam.port_name.get())
    nested_det.stats1.nd_array_port.put(nested_det.roi1.port_name.get())

    nested_det.validate_asyn_ports()

    tiff = nested_det.plugins.tiff1
    assert tiff is nested_det.get_plugin_by_asyn_port(tiff.port_name.get())


@pytest.mark.adsim
def test_visualize_asyn_digraph_smoke(ad_prefix, cleanup):
    # setup sim detector
    det = SimDetector(ad_prefix, name="test")
    cleanup.add(det)
    # smoke test
    det.visualize_asyn_digraph()


@pytest.mark.adsim
def test_read_configuration_smoke(ad_prefix, cleanup):
    class MyDetector(SingleTrigger, SimDetector):
        stats1 = Cpt(StatsPlugin, "Stats1:")
        proc1 = Cpt(ProcessPlugin, "Proc1:")
        roi1 = Cpt(ROIPlugin, "ROI1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.proc1.nd_array_port.put(det.cam.port_name.get())
    det.roi1.nd_array_port.put(det.proc1.port_name.get())
    det.stats1.nd_array_port.put(det.roi1.port_name.get())
    # smoke test
    det.stage()
    conf = det.stats1.read_configuration()
    desc = det.stats1.describe_configuration()
    det.unstage()
    for k in conf:
        assert k in desc

    assert len(conf) > 0
    assert len(conf) == len(desc)
    assert set(conf) == set(desc)


@pytest.mark.adsim
def test_str_smoke(ad_prefix, cleanup):
    class MyDetector(SingleTrigger, SimDetector):
        stats1 = Cpt(StatsPlugin, "Stats1:")
        proc1 = Cpt(ProcessPlugin, "Proc1:")
        roi1 = Cpt(ROIPlugin, "ROI1:")

    det = MyDetector(ad_prefix, name="test")
    cleanup.add(det)

    det.read_attrs = ["stats1"]
    det.stats1.read_attrs = ["mean_value"]
    det.stats1.mean_value.kind = Kind.hinted

    str(det)


@pytest.mark.adsim
def test_default_configuration_smoke(ad_prefix, cleanup):
    class MyDetector(SimDetector):
        imageplugin = Cpt(ImagePlugin, ImagePlugin._default_suffix)
        statsplugin = Cpt(StatsPlugin, StatsPlugin._default_suffix)
        colorconvplugin = Cpt(ColorConvPlugin, ColorConvPlugin._default_suffix)
        processplugin = Cpt(ProcessPlugin, ProcessPlugin._default_suffix)
        overlayplugin = Cpt(OverlayPlugin, OverlayPlugin._default_suffix)
        roiplugin = Cpt(ROIPlugin, ROIPlugin._default_suffix)
        transformplugin = Cpt(TransformPlugin, TransformPlugin._default_suffix)
        netcdfplugin = Cpt(NetCDFPlugin, NetCDFPlugin._default_suffix)
        tiffplugin = Cpt(TIFFPlugin, TIFFPlugin._default_suffix)
        jpegplugin = Cpt(JPEGPlugin, JPEGPlugin._default_suffix)
        # nexusplugin = Cpt(NexusPlugin, NexusPlugin._default_suffix)
        hdf5plugin = Cpt(HDF5Plugin, HDF5Plugin._default_suffix)
        # magickplugin = Cpt(MagickPlugin, MagickPlugin._default_suffix)

    d = MyDetector(ad_prefix, name="d")
    cleanup.add(d)
    d.stage()
    {n: getattr(d, n).read_configuration() for n in d.component_names}
    {n: getattr(d, n).describe_configuration() for n in d.component_names}
    d.unstage()


@pytest.mark.parametrize("plugin", _recursive_subclasses(PluginBase))
def test_default_configuration_attrs(plugin):
    configuration_attrs = plugin._default_configuration_attrs
    if configuration_attrs is None:
        pytest.skip("Configuration attrs unset")
    for k in configuration_attrs:
        assert hasattr(plugin, k)
        assert isinstance(getattr(plugin, k), (Component, DynamicDeviceComponent))


@pytest.fixture(scope="function")
def data_paths(request):
    def clean_dirs():
        shutil.rmtree("/tmp/ophyd_AD_test/data1")
        os.unlink("/tmp/ophyd_AD_test/data2")

    try:
        clean_dirs()
    except Exception:
        ...

    now = datetime.datetime.now()

    for year_offset in [-1, 0, 1]:
        make_dir_tree(now.year + year_offset, base_path="/tmp/ophyd_AD_test/data1")

    os.symlink("/tmp/ophyd_AD_test/data1", "/tmp/ophyd_AD_test/data2")
    request.addfinalizer(clean_dirs)


@pytest.mark.adsim
@pytest.mark.parametrize(
    "root,wpath,rpath,check_files",
    (
        (None, "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        (None, "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        ("/tmp/ophyd_AD_test/data1", "%Y/%m/%d", None, False),
        (
            "/tmp/ophyd_AD_test/data1",
            "/tmp/ophyd_AD_test/data1/%Y/%m/%d",
            "%Y/%m/%d",
            False,
        ),
        ("/", "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        (
            "/tmp/ophyd_AD_test/data2",
            "/tmp/ophyd_AD_test/data1/%Y/%m/%d",
            "%Y/%m/%d",
            True,
        ),
    ),
)
def test_fstiff_plugin(data_paths, ad_prefix, root, wpath, rpath, check_files, cleanup):
    fs = DummyFS()
    fs2 = DummyFS()
    if check_files:
        fh = pytest.importorskip("databroker.assets.handlers")

    class FS_tiff(TIFFPlugin, FileStoreTIFF, FileStoreIterativeWrite):
        pass

    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(
            FS_tiff,
            "TIFF1:",
            write_path_template=wpath,
            read_path_template=rpath,
            root=root,
            reg=fs,
        )

    target_root = root or "/"
    det = MyDetector(ad_prefix, name="det")
    cleanup.add(det)
    det.read_attrs = ["tiff1"]
    det.tiff1.read_attrs = []

    det.stage()
    st = det.trigger()
    while not st.done:
        time.sleep(0.1)
    reading = det.read()
    for name, d in det.collect_asset_docs():
        getattr(fs2, "insert_{name}".format(name=name))(**d)
    det.describe()
    det.unstage()

    datum_id = reading["det_image"]["value"]

    res_uid = fs.datum[datum_id]["resource"]
    res_doc = fs.resource[res_uid]
    assert res_doc == fs2.resource[res_uid]
    assert fs.datum[datum_id] == fs2.datum[datum_id]

    assert res_doc["root"] == target_root
    assert not PurePath(res_doc["resource_path"]).is_absolute()
    if check_files:
        time.sleep(5)  # Give AD some time to finish writing.
        path = PurePath(res_doc["root"]) / PurePath(res_doc["resource_path"])
        handler = fh.AreaDetectorTiffHandler(
            str(path) + os.path.sep, **res_doc["resource_kwargs"]
        )
        for fn in handler.get_file_list(
            datum["datum_kwargs"] for datum in fs.datum_by_resource[res_uid]
        ):
            assert Path(fn).exists()


@pytest.fixture
def h5py():
    try:
        import h5py
    except ImportError as ex:
        raise pytest.skip("h5py unavailable") from ex

    return h5py


@pytest.mark.adsim
@pytest.mark.parametrize(
    "root,wpath,rpath,check_files",
    (
        (None, "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        (None, "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        ("/tmp/ophyd_AD_test/data1", "%Y/%m/%d", None, False),
        (
            "/tmp/ophyd_AD_test/data1",
            "/tmp/ophyd_AD_test/data1/%Y/%m/%d",
            "%Y/%m/%d",
            False,
        ),
        ("/", "/tmp/ophyd_AD_test/data1/%Y/%m/%d", None, False),
        (
            "/tmp/ophyd_AD_test/data2",
            "/tmp/ophyd_AD_test/data1/%Y/%m/%d",
            "%Y/%m/%d",
            True,
        ),
    ),
)
def test_fshdf_plugin(
    h5py, data_paths, ad_prefix, root, wpath, rpath, check_files, cleanup
):
    fs = DummyFS()
    if check_files:
        fh = pytest.importorskip("databroker.assets.handlers")

    class FS_hdf(HDF5Plugin, FileStoreHDF5, FileStoreIterativeWrite):
        pass

    class MyDetector(SingleTrigger, SimDetector):
        hdf1 = Cpt(
            FS_hdf,
            "HDF1:",
            write_path_template=wpath,
            read_path_template=rpath,
            root=root,
            reg=fs,
        )

    target_root = root or "/"
    det = MyDetector(ad_prefix, name="det")
    cleanup.add(det)
    det.read_attrs = ["hdf1"]
    det.hdf1.read_attrs = []
    det.cam.acquire_time.put(0.1)
    det.cam.num_images.put(5)
    det.hdf1.warmup()
    time.sleep(3)

    det.stage()

    time.sleep(1)

    st = det.trigger()

    count = 0
    while not st.done:
        time.sleep(0.1)

        count += 1
        if count > 100:
            raise Exception("timedout")
    reading = det.read()
    det.describe()
    det.unstage()
    res_uid = fs.datum[reading["det_image"]["value"]]["resource"]
    res_doc = fs.resource[res_uid]
    assert res_doc["root"] == target_root
    assert not PurePath(res_doc["resource_path"]).is_absolute()
    if check_files:
        time.sleep(5)  # Give AD some time to finish writing.
        path = PurePath(res_doc["root"]) / PurePath(res_doc["resource_path"])
        handler = fh.AreaDetectorHDF5Handler(str(path), **res_doc["resource_kwargs"])
        for fn in handler.get_file_list(
            datum["datum_kwargs"] for datum in fs.datum_by_resource[res_uid]
        ):
            assert Path(fn).exists()


@pytest.mark.adsim
@pytest.mark.xfail
def test_many_connect(ad_prefix, cleanup):
    import ophyd

    pytest.skipif(
        ophyd.get_cl().name == "pyepics",
        "This is exposing race conditions in pyepics which " "cause segfaults.",
    )
    import gc

    fs = DummyFS()

    class FS_hdf(HDF5Plugin, FileStoreHDF5, FileStoreIterativeWrite):
        pass

    class MyDetector(SingleTrigger, SimDetector):
        hdf1 = Cpt(
            FS_hdf,
            "HDF1:",
            write_path_template="",
            read_path_template="",
            root="/",
            reg=fs,
        )

    try:
        from caproto.threading import client
    except ImportError:
        # caproto unavailable on python 3.5
        pass
    else:
        if client.SEARCH_MAX_DATAGRAM_BYTES > 1450:
            # old caproto compatibility - later versions lower to standardish
            # MTU-levels
            client.SEARCH_MAX_DATAGRAM_BYTES = 1450

    def tester():
        det = MyDetector(ad_prefix, name="det")
        print("made detector")
        try:
            print("*" * 25)
            print("about to murder socket")
            det.cam.acquire._read_pv._caproto_pv.circuit_manager._disconnected()
            print("murdered socket")
            print("*" * 25)
        except AttributeError:
            # must be pyepics
            pass
        det.destroy()
        del det
        gc.collect()

    for j in range(5):
        print(j)
        tester()


def test_ndderivedsignal_with_scalars():
    sig = Signal(value=np.zeros(12), name="zeros")
    shaped = NDDerivedSignal(sig, shape=(4, 3), num_dimensions=2, name="shaped")
    shaped.derived_shape == (4, 3)
    shaped.derived_ndims == 2
    assert shaped.get().shape == (4, 3)
    # Describe returns list
    assert shaped.describe()[shaped.name]["shape"] == [4, 3]
    shaped.put(np.ones((4, 3)))
    assert all(sig.get() == np.ones(12))


def test_ndderivedsignal_with_parent():
    class Detector(Device):
        flat_image = Component(Signal, value=np.ones(12))
        width = Component(Signal, value=4)
        height = Component(Signal, value=3)
        ndims = Component(Signal, value=2)
        shaped_image = Component(
            NDDerivedSignal,
            "flat_image",
            shape=("width", "height"),
            num_dimensions="ndims",
        )

    det = Detector(name="det")
    det.shaped_image.get().shape == (4, 3)
    cb = Mock()
    det.shaped_image.subscribe(cb)
    det.width.put(6)
    det.height.put(2)
    assert cb.called
    assert det.shaped_image._readback.shape == (6, 2)


@pytest.mark.adsim
@pytest.mark.parametrize("paths", [("/some/path/here"), ("/here/is/another/")])
def test_posix_path(paths, cleanup, ad_prefix):
    class MyDetector(SingleTrigger, SimDetector):
        tiff1 = Cpt(TIFFPlugin, "TIFF1:")

    det = MyDetector(ad_prefix, name="test")
    print(det.tiff1.plugin_type)
    cleanup.add(det)

    det.wait_for_connection()

    det.tiff1.file_path.put(paths)
    # det.cam.file_path.put(paths)
    det.stage()
    st = det.trigger()
    wait(st, timeout=5)
    det.unstage()



class OpException(Exception):
    """Ophyd base exception class"""

    pass


class ReadOnlyError(OpException):
    """Signal is read-only"""

    pass


class LimitError(ValueError, OpException):
    """Value is outside of defined limits"""

    pass


class DisconnectedError(OpException):
    """Signal or Device is not connected to EPICS"""

    pass


class DestroyedError(RuntimeError, DisconnectedError):
    """Signal or Device has been destroyed and is no longer usable"""

    pass


class ExceptionBundle(RuntimeError, OpException):
    """One or more exceptions was raised during a loop of try/except blocks"""

    def __init__(self, msg, exceptions):
        super().__init__(msg)
        self.exceptions = exceptions


class RedundantStaging(OpException):
    pass


class PluginMisconfigurationError(TypeError, OpException):
    # Keeping TypeError for backward-compatibility
    pass


class UnprimedPlugin(RuntimeError, OpException):
    ...


class UnknownStatusFailure(OpException):
    """
    Generic error when a Status object is marked success=False without details.
    """

    ...


class StatusTimeoutError(TimeoutError, OpException):
    """
    Timeout specified when a Status object was created has expired.
    """

    ...


class WaitTimeoutError(TimeoutError, OpException):
    """
    TimeoutError raised when we ware waiting on completion of a task.

    This is distinct from TimeoutError, just as concurrent.futures.TimeoutError
    is distinct from TimeoutError, to differentiate when the task itself has
    raised a TimeoutError.
    """

    ...


class InvalidState(RuntimeError, OpException):
    """
    When Status.set_finished() or Status.set_exception(exc) is called too late
    """

    ...



import logging
import pytest
import time

from unittest.mock import Mock
from ophyd.ophydobj import (
    OphydObject,
    register_instances_keyed_on_name,
    register_instances_in_weakset,
)
from ophyd.status import StatusBase, DeviceStatus, wait
from ophyd.utils import WaitTimeoutError

logger = logging.getLogger(__name__)


def test_status_basic():
    st = StatusBase()
    st.set_finished()


def test_status_callback_deprecated():
    "The old way, with finished_cb"
    st = StatusBase()
    cb = Mock()

    with pytest.warns(UserWarning):
        st.finished_cb = cb
    with pytest.warns(UserWarning):
        assert st.finished_cb is cb
    with pytest.raises(RuntimeError):
        st.finished_cb = None

    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    cb.assert_called_once_with(st)


def test_status_callback():
    "The new way, with add_callback and the callbacks property"
    st = StatusBase()
    cb = Mock()

    st.add_callback(cb)
    assert st.callbacks[0] is cb

    st.set_finished()
    st.wait(1)
    time.sleep(0.1)  # Wait for callbacks to run.
    cb.assert_called_once_with(st)


def test_status_others():
    DeviceStatus(None)


def test_status_wait():
    st = StatusBase()
    st.set_finished()
    wait(st)


def test_wait_status_failed():
    st = StatusBase(timeout=0.05)
    with pytest.raises(TimeoutError):
        wait(st)


def test_status_wait_timeout():
    st = StatusBase()

    with pytest.raises(WaitTimeoutError):
        wait(st, timeout=0.05)


def test_ophydobj():
    parent = OphydObject(name="name", parent=None)
    child = OphydObject(name="name", parent=parent)
    assert child.parent is parent

    with pytest.raises(ValueError):
        child.subscribe(None, event_type=None)

    with pytest.raises(KeyError):
        child.subscribe(lambda *args: None, event_type="unknown_event_type")

    assert parent.connected


def test_self_removing_cb():
    class TestObj(OphydObject):
        SUB_TEST = "value"

    test_obj = TestObj(name="name", parent=None)

    hit_A = 0
    hit_B = 0

    def remover(obj, **kwargs):
        nonlocal hit_A
        hit_A += 1
        obj.clear_sub(remover)

    def sitter(**kwargs):
        nonlocal hit_B
        hit_B += 1

    test_obj.subscribe(remover, "value")
    test_obj.subscribe(sitter, "value")
    test_obj._run_subs(sub_type="value")

    assert hit_A == 1
    assert hit_B == 1

    test_obj._run_subs(sub_type="value")

    assert hit_A == 1
    assert hit_B == 2


def test_unsubscribe():
    class TestObj(OphydObject):
        SUB_TEST = "value"

    test_obj = TestObj(name="name", parent=None)

    hit = 0

    def increment(**kwargs):
        nonlocal hit
        hit += 1

    cid = test_obj.subscribe(increment, "value")
    test_obj._run_subs(sub_type="value")
    assert hit == 1
    test_obj.unsubscribe(cid)
    test_obj._run_subs(sub_type="value")
    assert hit == 1

    # check multi unsubscribe
    test_obj.unsubscribe(cid)
    test_obj.clear_sub(increment)
    assert hit == 1


def test_unsubscribe_all():
    class TestObj(OphydObject):
        SUB_TEST = "value"

    test_obj = TestObj(name="name", parent=None)

    hit = 0

    def increment(**kwargs):
        nonlocal hit
        hit += 1

    test_obj.subscribe(increment, "value")
    test_obj._run_subs(sub_type="value")
    assert hit == 1
    test_obj.unsubscribe_all()
    test_obj._run_subs(sub_type="value")
    assert hit == 1


def test_subscribe_warn(recwarn):
    class TestObj(OphydObject):
        SUB_TEST = "value"
        _default_sub = SUB_TEST

    test_obj = TestObj(name="name", parent=None)
    test_obj.subscribe(lambda *args, **kwargs: None)
    assert len(recwarn) == 0


def test_subscribe_no_default():
    o = OphydObject(name="name", parent=None)

    with pytest.raises(ValueError):
        o.subscribe(lambda *a, **k: None)


def test_register_instance():
    weakset = register_instances_in_weakset()
    test1 = OphydObject(name="test1")
    assert test1 in weakset
    test2 = OphydObject(name="test1")
    assert test2 in weakset

    weakdict = register_instances_keyed_on_name()
    test1 = OphydObject(name="test1")
    assert weakdict["test1"] == test1
    test2 = OphydObject(name="test2")
    assert weakdict["test2"] == test2

    assert OphydObject._OphydObject__any_instantiated is True
    with pytest.raises(RuntimeError):
        register_instances_in_weakset(fail_if_late=True)
    with pytest.raises(RuntimeError):
        register_instances_keyed_on_name(fail_if_late=True)
